/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/* Access control functions */
static char *AtFSid = "$Header: acc.c[109.0] Wed Nov 24 03:47:01 1993 youki-k@is.aist-nara.ac.jp saved $";

#include <sys/file.h>		/* for R_OK */
#ifndef R_OK
#include <unistd.h>
#endif
#include <ctype.h>
#include <string.h>		/* for strchr */
#include "wfs.h"
#include "util.h"
#include "global.h"

/* all access control data is chained on acc_head with wf_accq */
typedef struct wf_accq {
	qelem	q;
	wf_acc	*accp;
	char	*dir;
} wf_accq;

extern qelem acc_head;
qelem acc_head = { &acc_head, &acc_head };

static wf_cap acc_cap[] = {
	{ "nfs", WF_ACC_NFS },
	{ "uip", WF_ACC_UIP },
	{ "fdp", WF_ACC_FDP },
	{ "raw", WF_ACC_RAW },
	{ NULL, 0 }
};

static unsigned long
wildcard_inet_addr(name, replacement)	/* in HOST byte order for later use */
char *name;
unsigned char replacement;
{
	unsigned long ip;
	char *p;
	int i;
	unsigned char *r;

	if (p = strchr(name, '*')) {
		/*
		 * IP address with wildcard; "163.221.*"
		 * We do not accept network number 0 for now, sorry.
		 */
		if (p == name) {
			/* star alone: class A to class C */
			if (replacement == 0)
				ip = 0;
			else /* if (replacement == 255) */
				ip = inet_addr("0xdf.0xff.0xff.0xff");
		} else {
			/*
			 * Ensure trailing dot is present.
			 *
			 * Probably tricky, undocumented feature:
			 * 133.1. --> 133.1.0.0
			 * 133.1  --> 133.0.0.1
			 * We use the former one.
			 */
			if (*(p-1) == '.') {
				*p = '\0';
			} else {
				*p = '.';
			}
			ip = inet_addr(name);
			*p = '*';
		}
	} else {
		ip = inet_addr(name);
	}

#ifndef	INADDR_NONE
#define	INADDR_NONE	0xffffffff
#endif
	if ((ip & INADDR_NONE) == INADDR_NONE) {
		dlog("Invalid address specification: %s", name);
	} else {
		r = (char *) &ip;
		for (i = 0; i < sizeof(ip); ++i) {
			if (r[i] == 0) r[i] = replacement;
		}
		ip = ntohl(ip);
	}
	return ip;
}

static wf_acc *new_accp;
static wf_accq *new_accq;

int
acc_bind(name, value)
char *name;
char *value;
{
	char *p;
	wf_acc *accp;
	unsigned long from, to;
	long flag;

	if (isdigit(name[0])) {
		if (p = strchr(name, '-')) {
			*p = '\0';
			from = wildcard_inet_addr(name, 0);
			to = wildcard_inet_addr(p+1, 255);
			*p = '-';
		} else {
			from = wildcard_inet_addr(name, 0);
			to = wildcard_inet_addr(name, 255);
		}
		if (from == INADDR_NONE || to == INADDR_NONE) {
			dlog("invalid IP address");
			return -1;
		}
		flag = parse_cap(acc_cap, value);
		if (flag < 0) {
			dlog("invalid flag");
			return -1;
		}
	} else {
		/*
		 * Domain name or special keywords?
		 * Just reject them for now.
		 */
		dlog("Not an IP address");
		return -1;
	}
	accp = ALLOC(wf_acc);
	accp->from = from;
	accp->to = to;
	accp->flag = flag;
	q_insert(accp, new_accp);
	return 0;
}

wf_acc *
acc_load(dir)
char *dir;
{
	char path[MAXPATHLEN];

	new_accp = (wf_acc *) q_alloc();
	sprintf(path, "%s/vol/%s/access", cs_topdir, dir);
	if (access(path, F_OK) < 0 || parse_conf(path, acc_bind) < 0) {
		FREE(new_accp);
		return (wf_acc *)0;
	}
	new_accq = ALLOC(wf_accq);
	new_accq->accp = new_accp;
	new_accq->dir = strdup(dir);
	q_insert(new_accq, LAST(wf_accq, &acc_head));
	return new_accp;
}

/* find access control data by directory name */
wf_acc *
acc_findname(dir)
char *dir;
{
	wf_accq *accq;

	/* search memory cache */
	ITER(accq, wf_accq, &acc_head) {
		if (strcmp(accq->dir, dir) == 0) {
			return accq->accp;
		}
	}

	/* search disk cache */
	return acc_load(dir);
}

int
acc_check(accp, ip, mask)
wf_acc *accp;
unsigned long ip;
long mask;
{
	unsigned long host_ip;
	wf_acc *p;

	host_ip = ntohl(ip);
	ITER(p, wf_acc, &accp->q) {
		/* comparison in HOST byte order */
		if (p->from <= host_ip && host_ip <= p->to
		    && (p->flag & mask)) {
			return 0;
		}
	}
	return -1;
}

