/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
static char *AtFSid = "$Header: file.c[109.1] Tue Dec  7 23:37:58 1993 youki-k@is.aist-nara.ac.jp saved $";

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>		/* for R_OK */
#ifndef R_OK
#include <unistd.h>
#endif
#include <fcntl.h>		/* for O_WRONLY etc. */
#include "wfs.h"
#include "util.h"
#include "global.h"

static
unsigned int type2mode[NFFIFO+1] = {	/* depends on enum ftype */
	0,		/* NFNON */
	S_IFREG,
	S_IFDIR,
	S_IFBLK,
	S_IFCHR,
	S_IFLNK,
	S_IFSOCK,
	0,		/* NFBAD */
	S_IFIFO
};

#define	WF_FILE_HASHSIZE	16

typedef struct fcache {
		wf_file *filep;
		wf_dir	*dirp;
} fcache;
static fcache findid_cache[WF_FILE_HASHSIZE],
	findname_cache[WF_FILE_HASHSIZE];

wf_file *
file_alloc()
{
	wf_file *filep;
	filep = ALLOC(wf_file);
	return filep;
}

void
file_free(filep)
wf_file *filep;
{
	if (filep->name) free(filep->name);
	if (filep->link) free(filep->link);
	FREE(filep);
}

void
file_uncache(dirp)
wf_dir *dirp;
{
	int i = dirp->id % WF_FILE_HASHSIZE;
	fcache *p;

	p = &findid_cache[i];
	if (p->dirp == dirp) {
		p->dirp = (wf_dir *) 0;
		p->filep = (wf_file *) 0;
	}
	p = &findname_cache[i];
	if (p->dirp == dirp) {
		p->dirp = (wf_dir *) 0;
		p->filep = (wf_file *) 0;
	}
}

wf_file *
file_new(name, link, filetype, nlink, mode, uid, gid, size, time, fileid, fsid)
char *name, *link;
ftype filetype;
u_int nlink, mode, uid, gid, fsid;
long size, time, fileid;
{
	wf_file *filep = ALLOC(wf_file);
	filep->name = name;
	filep->crc = updcrc(0L, filep->name, strlen(filep->name)+1);
	filep->link = link;
	if (fileid == 0) {
		filep->id = filep->fattr.fileid = filep->crc;
	} else {
		filep->id = filep->fattr.fileid = fileid;
	}

	filep->fattr.type = filetype;
	filep->fattr.mode = mode;
	filep->fattr.mode |= type2mode[filetype];
	filep->fattr.nlink = nlink;
	filep->fattr.uid = uid;
	filep->fattr.gid = gid;
	filep->fattr.size = size;
	filep->fattr.blocksize = 8192;
	filep->fattr.blocks = compute_blocks(filep->fattr.size);
	filep->fattr.fsid = fsid;
	filep->fattr.atime.seconds = filep->fattr.mtime.seconds
		= filep->fattr.ctime.seconds = time;
	return filep;
}

void
attr_initdir(attrp)
fattr *attrp;
{
	attrp->nlink = 2;		/* should be updated later */
	attrp->type = NFDIR;
	attrp->mode = 0755;
	attrp->mode |= type2mode[attrp->type];
	attrp->size = 512;
	attrp->blocksize = 8192;
	attrp->blocks = compute_blocks(attrp->size);
}

void
attr_initfile(attrp, size)
fattr *attrp;
long size;
{
	attrp->type = NFREG;
	attrp->mode = 0644;
	attrp->mode |= type2mode[attrp->type];
	attrp->nlink = 1;
	attrp->size = size;
	attrp->blocksize = 8192;
	attrp->blocks = compute_blocks(attrp->size);
}

unsigned int
attr_touch(attrp)
fattr *attrp;
{
	return attrp->atime.seconds = attrp->mtime.seconds = gettime();
}

wf_file *
file_newvol(name)
char *name;
{
	wf_file *filep;
	fattr *attrp;

	filep = file_alloc();
	filep->name = strdup(name);
	filep->crc = updcrc(0L, name, strlen(name)+1);
	attrp = &filep->fattr;
	attr_initdir(attrp);
	attr_touch(attrp);
	return filep;
}

/* find file by (filesystem id, file id). */
wf_file *
file_findid(dirp, fsid, fileid)
wf_dir *dirp;
long fsid, fileid;
{
	wf_file *filep;
	qelem *q_head = &dirp->file->q;
	fcache *p;

	/*
	 * lookup cache
	 */
	/* getattr operations are likely to be performed sequentially. */
	p = &findid_cache[dirp->id % WF_FILE_HASHSIZE];
	if (p->filep && p->dirp == dirp) {
		filep = NEXT(wf_file, p->filep);
		if (filep != HEAD(wf_file, q_head)
		    && FID(filep) == fileid && FSID(filep) == fsid) {
			p->filep = filep;
			return filep;
		} else {
			p->filep = (wf_file *) 0;	/* stale */
		}
	}

	/*
	 * resort to exhaustive search
	 */
	ITER(filep, wf_file, q_head) {
		if (FID(filep) == fileid && FSID(filep) == fsid) {
			p->filep = filep;
			p->dirp = dirp;
			return filep;
		}
	}
	return NULL;
}

/* find file by filename. */
wf_file *
file_findname(dirp, name)
wf_dir *dirp;
char *name;
{
	wf_file *filep;
	qelem *q_head = &dirp->file->q;
	u_int crc;
	fcache *p;

	/*
	 * lookup cache
	 */
	/* lookup operations are likely to be performed sequentially.
	 * we declare another hashtable for fine tuning.
	 */
	p = &findname_cache[dirp->id % WF_FILE_HASHSIZE];
	if (p->filep && p->dirp == dirp) {
		filep = NEXT(wf_file, p->filep);
		if (filep != HEAD(wf_file, q_head)
		    && strcmp(filep->name, name) == 0) {
			p->filep = filep;
			return filep;
		} else {
			p->filep = (wf_file *) 0;	/* stale */
		}
	}

	/*
	 * resort to exhaustive search
	 */
	crc = updcrc(0L, name, strlen(name)+1);
	ITER(filep, wf_file, q_head) {
		if (crc == filep->crc
		    && strcmp(filep->name, name) == 0) {
			p->filep = filep;
			p->dirp = dirp;
			return filep;
		}
	}
	return NULL;
}

int
file_recv_init(volp, dirp, filep)
wf_vol *volp;
wf_dir *dirp;
wf_file *filep;
{
	int fd;
	char path[MAXPATHLEN];

	sprintf(path, "%s/ongoing/xfer.%lx.%lx.%s", cs_topdir,
		volp->id, dirp->id, filep->name);
	fd = open(path, O_WRONLY|O_CREAT|O_TRUNC, 0660);
	if (fd < 0)
		dlog("ftp_recv_init: couldn't create file");
	return fd;
}

void
file_recv_finish(volp, dirp, filep)
wf_vol *volp;
wf_dir *dirp;
wf_file *filep;
{
	char path[MAXPATHLEN], path2[MAXPATHLEN];

	sprintf(path, "%s/ongoing/xfer.%lx.%lx.%s", cs_topdir,
		volp->id, dirp->id, filep->name);
	file_getpath(path2, volp, dirp, filep);
	if (mkdirs(path2))
		dir_realdir(path2, volp, dirp);
	rename(path, path2);
}

void
file_recv_abort(volp, dirp, filep)
wf_vol *volp;
wf_dir *dirp;
wf_file *filep;
{
	char path[MAXPATHLEN], path2[MAXPATHLEN];

	sprintf(path, "%s/ongoing/xfer.%lx.%lx.%s", cs_topdir,
		volp->id, dirp->id, filep->name);
	sprintf(path2, "%s/stale/%s.%d/xfer.%lx.%lx.%s",
		cs_topdir, cs_date, cs_serial,
		volp->id, dirp->id, filep->name);
	mkdirs(path2);
	rename(path, path2);
}

void
file_getpath(buf, volp, dirp, filep)
char *buf;
wf_vol *volp;
wf_dir *dirp;
wf_file *filep;
{
	if (volp->flag & WF_VOL_DIRECTFILE) {
		sprintf(buf, "%s/%s/%s",
			volp->topdir, dirp->name, filep->name);
	} else {
		sprintf(buf, "%s/cache/%lx/%lx/%s",
			cs_topdir, volp->id, dirp->id, filep->name);
	}
}

/* see if file is cached. */
bool_t
file_findcache(volp, dirp, filep)
wf_vol *volp;
wf_dir *dirp;
wf_file *filep;
{
	struct stat s;
	char path[MAXPATHLEN];

	file_getpath(path, volp, dirp, filep);
	if (stat(path, &s) < 0)
		return -1;
	if (! (s.st_mode & S_IREAD) || (s.st_mode & S_IFMT) != S_IFREG)
		return -1;
	if (! (filep->flag & WF_FILE_MUTABLE)
	    && s.st_size != filep->fattr.size) {
		/* XXX: update policy */
		file_dispose(volp, dirp, filep);
		return -1;
	}
	return 0;
}

/* make that file stale. */
void
file_dispose(volp, dirp, filep)
wf_vol *volp;
wf_dir *dirp;
wf_file *filep;
{
	char path[MAXPATHLEN], path2[MAXPATHLEN];

	file_getpath(path, volp, dirp, filep);
	sprintf(path2, "%s/stale/%s.%d/file.%lx.%lx.%s",
		cs_topdir, cs_date, cs_serial,
		volp->id, dirp->id, filep->name);
	mkdirs(path2);
	rename(path, path2);
}

/* WWUIP DIAG */
void
file_inspect(reqp, name)
wf_req *reqp;
char *name;
{
	int id = reqp->id;
	int ret;
	wf_ih ih;

	ret = root_name2ih(name, &ih);
	if (ret != WF_REP_OK) {
		req_send(reqp, "513.%03d root_name2ih: error code %d.\n",
			 id, ret);
		return;
	}
	req_send(reqp, "220-%03d fil %s\n", id, name);
	req_send(reqp, "213-%03d name = %s\n", id, PROT(ih.filep->name));
	req_send(reqp, "213-%03d link = %s\n", id, PROT(ih.filep->link));
	req_send(reqp, "213-%03d id = %lx\n", id, ih.filep->id);
	req_send(reqp, "213-%03d crc = %lx\n", id, ih.filep->crc);
	req_send(reqp, "213-%03d flag = %x\n", id, ih.filep->flag);
	req_send(reqp, "213-%03d type = %lx\n", id, ih.filep->fattr.type);
	req_send(reqp, "213-%03d fsid = %lx\n", id, ih.filep->fattr.fsid);
	req_send(reqp, "213.%03d fileid = %lx\n", id, ih.filep->fattr.fileid);
}

