/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/* NFS over FTP */
static char *AtFSid = "$Header: ftp_retr.c[109.0] Wed Nov 24 03:47:10 1993 youki-k@is.aist-nara.ac.jp saved $";

#include <sys/ioctl.h>		/* for FIONREAD */
#include "wfs.h"
#include "util.h"
#include "global.h"

#define RCV_BUFSIZE 32768	/* larger is better on giganets */
char cs_databuf[RCV_BUFSIZE];

static void ftp_recv_retr();

static void
ftp_read_lock(c)
wf_thrd *c;
{
	c->filep->thrdp = c;
	++c->dirp->n_xfer;
}

static void
ftp_read_unlock(c)
wf_thrd *c;
{
	warn_negative(--c->dirp->n_xfer);
	c->filep->thrdp = (wf_thrd *)0;
	if (c->cp)
		c->cp->thrdp = (wf_thrd *)0;
}

static void ftp_read_2(), ftp_read_3(), ftp_read_timeout(), ftp_read_fail();
static void ftp_port_close();
static int ftp_port_open(), ftp_port_accept(), ftp_send_port();

void
ftp_read_miss(c)
wf_thrd *c;
/* in: rqstp, volp, offset, count, totalcount, dirp, filep */
/* out: cp, flag */
{
	ftp_read_lock(c);
	c->callout =
		timeout_set(gettime()+1, ftp_read_timeout, c);
	c->cp = vol_getconn(c->child_volp);
	if (c->cp == (wf_conn *)0) {
		if ((c->child_volp->flag & WF_VOL_XCONN)
		    || (c->cp = vol_newconn(c->child_volp)) == (wf_conn *)0) {
			/* exceeded resource limit */
			ftp_read_unlock(c);
			timeout_clear(c->callout);
			(*c->reply)(c, WF_ERR_NQUOT);
			return;
		}
		assert(c->cp != (wf_conn *)0);
		/* new connection needed */
		sched_task(ftp_read_2, c, &c->cp);
		thrd_watch(c);
		c->flag |= WF_THREAD_BUSY;
		ftp_conn_open(c);
		return;
	}
	c->flag |= WF_THREAD_BUSY;
	ftp_read_2(c);
}

static void
ftp_retr_status(c)
wf_thrd *c;
{
	int ret;

	ret = ftp_recv_status(c->cp);
	if (c->flag & WF_THREAD_ERROR) {
		if (ret == 426)
			ftp_read_fail(c);
		else
			ftp_read_3(c);
	}
}

static void
ftp_read_2(c)
wf_thrd *c;
{
	int ret;

	c->idle = 0;
	switch (c->step) {
	case FTPS_CONN_FAIL:
		ftp_read_fail(c);
		return;
	case FTPS_PROMPT:
#ifdef VERBOSE
		talk(c->client, "retrieving file %s/%s",
		     c->dirp->name, c->filep->name);
		dlog("retrieving file %s/%s", c->dirp->name, c->filep->name);
#endif
		c->cp->thrdp = c;
		c->flag |= WF_THREAD_RISKY;
		c->give_up = ftp_read_fail;
		c->closure = (void *)c;
		ftp_port_open(c);
		ret = ftp_send_port(c);
		if (ret < 0) {
			ftp_read_fail(c);
			return;
		}
		c->step = FTPS_RETR_SEND;
		so_clear(c->cp->id);
		so_callback(c->cp->id, ftp_read_2, c);
		return;
	case FTPS_RETR_SEND:
		ret = ftp_recv_status(c->cp);
		if (ret == 200) {
			so_clear(c->cp->id);
			so_callback(c->cp->id, ftp_retr_status, c);
			tcp_send(c->cp->id, "RETR /%s%s/%s\r\n",
				 c->cp->info, c->dirp->name, c->filep->name);
			if (cs_ftp_debug) {
				req_send(cs_ftp_debug, "RETR /%s%s/%s\r\n",
					 c->cp->info, c->dirp->name,
					 c->filep->name);
			}
			c->step = FTPS_PORT_ACCEPT;
			so_callback(c->so, ftp_read_2, c);
		} else if (ret == 0 || ret < 400) {
			/* expect more replies */
			c->step = FTPS_RETR_SEND;
		} else {
			ftp_read_3(c);
		}
		return;
	case FTPS_PORT_ACCEPT:
		so_clear(c->so);
		ftp_port_accept(c);
		/* ready for data transfer. now receive data. */
		c->step = FTPS_RETR_RECVINIT;
		sched_task(ftp_read_3, c, &c->so);
		so_callback(c->so, ftp_recv_retr, c);
		return;
	}
}

#ifdef TRACE
static void
ftp_retr_trace(c)
wf_thrd *c;
{
	wf_trace *tracep;
	tracep = ALLOC(wf_trace);
	tracep->time = gettime();
	tracep->volume = c->child_volp->id;
	tracep->dir = c->dirp->id;
	tracep->file = c->filep->id;
	tracep->op = WF_OP_READ;
	tracep->hit = WF_CACHE_READY;
	tracep->arg = (long) c->filep->fattr.size;
	trace_event(c, tracep);
}
#endif

static void
ftp_read_3(c)
wf_thrd *c;
{
	int ret;

	so_clear(c->cp->id);
	so_callback(c->cp->id, ftp_recv_junk, (void *)c->cp->id);
	ftp_port_close(c->so);
	ftp_read_unlock(c);
	vol_releaseconn(c->child_volp);
	if (c->flag & WF_THREAD_ERROR) {
#ifdef VERBOSE
		talk(c->client, "couldn't retrieve %s/%s",
		     c->dirp->name, c->filep->name);
		dlog("couldn't retrieve %s/%s",
		     c->dirp->name, c->filep->name);
#endif
	} else {
		file_recv_finish(c->child_volp, c->dirp, c->filep);
#ifdef VERBOSE
		talk(c->client, "retrieved %s/%s",
		     c->dirp->name, c->filep->name);
		dlog("retrieved %s/%s", c->dirp->name, c->filep->name);
#endif
	}
	if (! (c->flag & WF_THREAD_TIMEOUT)) {
		timeout_clear(c->callout);
	}
	c->flag &= ~WF_THREAD_BUSY;
	if (c->flag & WF_THREAD_ERROR) {
		(*c->reply)(c, WF_ERR_IO);
	} else {
#ifdef TRACE
		ftp_retr_trace(c);
#endif
		(*c->reply)(c, WF_REP_OK);
	}
}

static void
ftp_read_timeout(c)
wf_thrd *c;
{
	c->flag |= WF_THREAD_TIMEOUT;
	(*c->reply)(c, WF_REP_WAIT);
}

static void
ftp_read_fail(c)
wf_thrd *c;
{
	wf_conn *cp = c->cp;
	c->flag |= WF_THREAD_ERROR;
	ftp_read_3(c);
	conn_shutdown(cp, 2);
}

static int
ftp_port_open(c)
wf_thrd *c;
{
	int so, opt;
	struct sockaddr_in sin;
	int sinlen;
	int ret;

	so = socket(AF_INET, SOCK_STREAM, 0);
	if (so < 0) {
		errno_diag();
		return -1;
	}
	opt = 1;
	ioctl(so, FIONBIO, &opt);
	setsockopt(so, SOL_SOCKET, SO_REUSEADDR, (char *)&opt, sizeof(int));
	opt = RCV_BUFSIZE;
	setsockopt(so, SOL_SOCKET, SO_RCVBUF, (char *)&opt, sizeof(int));
	so_register(so, WF_SO_READ);
	bzero(&sin, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl(INADDR_ANY);
	ret = bind(so, (struct sockaddr *) &sin, sizeof (sin));
	if (ret < 0) {
		errno_diag();
		return ret;
	}
	sinlen = sizeof (sin);
	getsockname(so, (struct sockaddr *) &sin, &sinlen);
	c->port = ntohs(sin.sin_port);
	c->so = so;
	listen(so, 1);
	return ret;
}

static int
ftp_port_accept(c)
wf_thrd *c;
{
	int ret;
	sockaddr_in sin;
	int sinlen;

	sinlen = sizeof(sin);
	ret = accept(c->so, (struct sockaddr *) &sin, &sinlen);
	if (ret < 0) {
		errno_diag();
		return ret;
	}
	close(c->so);
	so_unregister(c->so, WF_SO_READ);
	c->so = ret;
	so_register(ret, WF_SO_READ);
	if (bcmp(&sin.sin_addr, &c->cp->srv->sin.sin_addr,
		 sizeof (struct in_addr))) {
		syslog(LOG_WARNING, "ftp_port_accept: sin_addr mismatch");
		return -1;
	}
	return ret;
}

static void
ftp_port_close(so)
int so;
{
	so_clear(so);
	so_unregister(so, WF_SO_ALL);
	close(so);
}

static int
ftp_send_port(c)
wf_thrd *c;
{
	char buf[64];
	int ret, sinlen;
	long ipaddr;
	sockaddr_in sin;

	sinlen = sizeof (sin);
	getsockname(c->cp->id, (struct sockaddr *)&sin, &sinlen);
	/* ipforwarding fix from utashiro@sramhc.sra.co.jp */
	ipaddr = ntohl(sin.sin_addr.s_addr);
	ret = tcp_send(c->cp->id, "PORT %d,%d,%d,%d,%d,%d\r\n", 
		       ((ipaddr >> 24) & 0xFF),
		       ((ipaddr >> 16) & 0xFF),
		       ((ipaddr >> 8) & 0xFF),
		       ((ipaddr >> 0) & 0xFF),
		       ((c->port >> 8) & 0xFF),
		       ((c->port >> 0) & 0xFF));
	if (cs_ftp_debug) {
		req_send(cs_ftp_debug, "PORT %d,%d,%d,%d,%d,%d\r\n", 
			 ((ipaddr >> 24) & 0xFF),
			 ((ipaddr >> 16) & 0xFF),
			 ((ipaddr >> 8) & 0xFF),
			 ((ipaddr >> 0) & 0xFF),
			 ((c->port >> 8) & 0xFF),
			 ((c->port >> 0) & 0xFF));
	}
	return ret;
}

static void
ftp_recv_retr(c)
wf_thrd *c;
{
	long navail;
	int nread;
	int ret, optlen;

	if (c->flag & WF_THREAD_ERROR) {
#ifdef VERBOSE
		dlog("transfer error, discarding %s", c->filep->name);
#endif
		close(c->fd);
		file_recv_abort(c->child_volp, c->dirp, c->filep);
		c->step = FTPS_RETR_FIN;
	}
	c->idle = 0;	/* Hey, I'm alive and healthy */
	switch (c->step) {
	case FTPS_RETR_RECVINIT:
		c->fd = file_recv_init(c->child_volp, c->dirp, c->filep);
		c->step = FTPS_RETR_RECVLOOP;
	case FTPS_RETR_RECVLOOP:
		ioctl(c->so, FIONREAD, &navail);
		nread = MIN(navail, RCV_BUFSIZE);
		if (nread == 0) {
			optlen = sizeof(int);
			getsockopt(c->so, SOL_SOCKET, SO_ERROR,
				   &ret, &optlen);
			if (ret && ret < sys_nerr) {
				dlog("ftp_recv_retr: %s", sys_errlist[ret]);
			}
			/* otherwise, transfer completed. close anyway */
			close(c->fd);
			break;
		}
		ret = recv(c->so, cs_databuf, nread, 0);
		if (ret < 0) {
			errno_diag();
			c->flag |= WF_THREAD_ERROR;
			close(c->fd);
			break;
		}
		c->cp->stats.bytes_from_server += nread;
		c->child_volp->stats.bytes_from_server += nread;
		ret = write(c->fd, cs_databuf, nread);
		if (ret < 0) {
			errno_diag();
			c->flag |= WF_THREAD_ERROR;
			close(c->fd);
			break;
		}
		return;
	case FTPS_RETR_FIN:
		break;
	}
	wakeup(&c->so);
}
