/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
static char *AtFSid = "$Header: icmp.c[109.0] Wed Nov 24 03:47:10 1993 youki-k@is.aist-nara.ac.jp saved $";

#include <sys/types.h>
#include <sys/param.h>
#include "wfs.h"
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netdb.h>
#include "util.h"
#include "global.h"

static int seq;
static int so;
#define	WF_ICMP_LEN	56
#define	WF_ICMP_MAX	512	/* XXX should be enough */
static char buf[WF_ICMP_MAX];

void
icmp_start()
{
	struct protoent *proto;
	proto = getprotobyname("icmp");
	so = socket(AF_INET, SOCK_RAW, proto->p_proto);
	if (so < 0) {
		errno_diag();
		return;
	}
	so_register(so, WF_SO_READ);
	so_callback(so, icmp_recv, (void *)so);
}

static int
icmp_cksum(ptr, nbytes)
u_short *ptr;
int nbytes;
{
	/* icmp_cksum is verbatim copy, whilst I checked errata carefully, from
	 * "UNIX Network Programming", W. Richard Stevens
	 *
	 * Refer to "Computing the Internet Checksum" by R.Braden, D.Borman and
	 * C.Partridge, Computer Communication Review 19(2), April 1989,
	 * pp.86-101, for additional details.
	 */
	long sum;
	u_short oddbyte;
	u_short answer;

	sum = 0;
	while (nbytes > 1) {
		sum += *ptr++;
		nbytes -= 2;
	}
	if (nbytes == 1) {
		oddbyte = 0;
		*((u_char *) &oddbyte) = *(u_char *)ptr;
		sum += oddbyte;
	}
	sum = (sum >> 16) + (sum & 0xffff);
	sum += (sum >> 16);
	answer = ~sum;
	return answer;
}

void
icmp_send(to)
sockaddr_in *to;
{
	struct icmp *p = (struct icmp *)buf;

	p->icmp_type = ICMP_ECHO;
	p->icmp_code = 0;
	p->icmp_cksum = 0;
	p->icmp_id = cs_pid & 0xFFFF;
	p->icmp_seq = seq++;
	gettimeofday((struct timeval *) buf+ICMP_MINLEN, (struct timezone *)0);
	p->icmp_cksum = icmp_cksum(p, WF_ICMP_LEN);

	sendto(so, buf, WF_ICMP_LEN, 0, (struct sockaddr *)to, sizeof(*to));
}

static void
tvsub(out, in)
struct timeval *out, *in;
{
	/* also from "UNIX Network Programming", W. Richard Stevens */
	if ((out->tv_usec -= in->tv_usec) < 0) {
		out->tv_sec--;
		out->tv_usec += 1000000;
	}
	out->tv_sec -= in->tv_sec;
}

void
icmp_recv(so)
int so;
{
	int n, fromlen;
	struct sockaddr_in from;
	int iphdrlen;
	struct ip *ip = (struct ip *)buf;
	wf_srv *srvp;
	wf_srvstat *stat;
	struct icmp *p;
	struct timeval tv;
	long rtt;		/* in milliseconds */

	/*
	 * receive packet
	 */
	fromlen = sizeof(from);
 resume:
	n = recvfrom(so, buf, WF_ICMP_MAX, 0, (struct sockaddr *)&from,
		       &fromlen);
	if (n < 0) {
		errno_diag();
		return;
	}

	/*
	 * verify header
	 */
	iphdrlen = ip->ip_hl << 2;
	if (n < iphdrlen + ICMP_MINLEN) {
		dlog("icmp: packet too short");
		return;
	}
	n -= iphdrlen;

	/*
	 * analyse
	 */
	p = (struct icmp *) (buf + iphdrlen);
	switch(p->icmp_type) {
	case ICMP_ECHOREPLY:
		/* check if it's response to my echo request */
		if (p->icmp_id != (cs_pid & 0xFFFF))
			return;
		/* check if I have talked with the server */
		srvp = srv_findip(from.sin_addr.s_addr);
		if (! srvp)
			return;
		stat = &srvp->stats;
		gettimeofday((struct timeval *)&tv, (struct timezone *) 0);
		tvsub(&tv, (struct timeval *) p->icmp_data);
		rtt = tv.tv_sec * 1000 + (tv.tv_usec / 1000);
		stat->icmp_reply++;
		if (stat->rtt)
			stat->rtt = stat->rtt * 3 / 4 + rtt / 4;
		else
			stat->rtt = rtt;	/* initial boost */
		return;
	case ICMP_UNREACH:
		switch(p->icmp_code) {
		case 2:		/* ICMP_UNREACH_PROTOCOL */
			dlog("%s: bad protocol", inet_ntoa(from.sin_addr));
			return;
		case 3:		/* ICMP_UNREACH_PORT */
#if 0	/* it's noisy... */
			dlog("%s: bad port", inet_ntoa(from.sin_addr));
#endif
			return;
		case 4:		/* ICMP_UNREACH_NEEDFRAG */
			dlog("%s: IP_DF caused drop",
			     inet_ntoa(from.sin_addr));
			return;
		case 5:		/* ICMP_UNREACH_SRCFAIL */
			dlog("%s: source route failed",
			     inet_ntoa(from.sin_addr));
			return;
		}
		ip = (struct ip *) (buf + iphdrlen + 8);
		/* check if it's response to my packet */
		if (ip->ip_src.s_addr != cs_ipaddr.s_addr)
			return;
		/* check if I have talked with the server */
		srvp = srv_findip(ip->ip_dst.s_addr);
		if (! srvp)
			return;
		switch(p->icmp_code) {
		case 0:		/* ICMP_UNREACH_NET */
			dlog("%s: network unreachable", srvp->name);
			srv_flag(srvp, WF_SRV_NUNREACH);
			thrd_srvfailure(srvp);
			return;
		case 1:		/* ICMP_UNREACH_HOST */
			dlog("%s: host unreachable", srvp->name);
			srv_flag(srvp, WF_SRV_HUNREACH);
			thrd_srvfailure(srvp);
			return;
		}
		return;
	case ICMP_SOURCEQUENCH:
		return;
	case ICMP_TIMXCEED:
		return;
	case ICMP_PARAMPROB:
		return;
	default:
#if 0	/* it's noisy... */
		dlog("spurious ICMP packet from %s, type=%d",
		     inet_ntoa(from.sin_addr), p->icmp_type);
#endif
		return;
	}
}

