/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/* volume: caching, file io */
static char *AtFSid = "$Header: vol.c[109.1] Tue Nov 30 02:47:12 1993 youki-k@is.aist-nara.ac.jp saved $";

#include <sys/file.h>		/* for R_OK */
#ifndef R_OK
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include "wfs.h"
#include "util.h"
#include "global.h"

extern qelem vol_head;
qelem vol_head = { &vol_head, &vol_head };

wf_vol *
vol_alloc()
{
	wf_vol *volp;
	volp = ALLOC(wf_vol);
	volp->srvq = (wf_srvq *) q_alloc();
	volp->dirp = (wf_dir *) q_alloc();
	q_insert(volp, &vol_head);
	return volp;
}

void
vol_free(volp)
wf_vol *volp;
{
	if (volp->id == 0) {
		syslog(LOG_WARNING, "volp->id == 0");
	}
	free(volp->dir);
	FREE(volp->srvq);
	FREE(volp->dirp);
	q_remove(volp);
	FREE(volp);
}

/* find volume by id */
wf_vol *
vol_findid(id)
long id;
{
	wf_vol *volp;

	/* search memory cache */
	ITER(volp, wf_vol, &vol_head) {
		if (volp->id == id) {
			return volp;
		}
	}
	return NULL;
}

/* find volume by name */
wf_vol *
vol_findname(dir, name)
char *dir, *name;
{
	wf_vol *volp;

	/* search memory cache */
	ITER(volp, wf_vol, &vol_head) {
		if (strcmp(volp->dir, dir) == 0
		    && strcmp(volp->name, name) == 0) {
			return volp;
		}
	}

	/* search disk cache */
	volp = vol_load(dir, name);
	return volp;
}

static wf_vol *new_volp;
static wf_srv *new_srvp;
static int dir_count, srv_count;

static int
vol_set_id(id)
char *id;
{
	new_volp->id = atoi(id);
	if (new_volp->id == WF_ROOT_VOL_ID) {
		dlog("incorrect volume id");
		return -1;
	}
	return 0;
}

int
vol_set_name(name)
char *name;
{
	dlog("obsolete entry: volume-name: ignored");
	return 0;
}

int
vol_set_ftp_server(hostname)
char *hostname;
{
	if (srv_count > dir_count + 1) {
		dlog("incorrect server-directory sequence");
		return -1;
	}
	new_srvp = srv_findname(hostname);
	if (new_srvp == (wf_srv *)0) {
		return -1;
	}
	++srv_count;
	return 0;
}

int
vol_set_ftp_directory(subdir)
char *subdir;
{
	if (!new_srvp || dir_count > srv_count + 1) {
		dlog("incorrect server-directory sequence");
		return -1;
	}
	vol_addsrv(new_volp, new_srvp, subdir);
	++dir_count;
	return 0;
}

static wf_cap ftp_cap[] = {
	{ "follow", WF_FTP_FOLLOW },
	{ "nostat", WF_FTP_NOSTAT },
	{ NULL, 0 }
};

int
vol_set_ftp_options(flagstr)
char *flagstr;
{
	long flag;

	if (!new_srvp) {
		dlog("misplaced ftp-options");
		return -1;
	}
	flag = parse_cap(ftp_cap, flagstr);
	if (flag < 0) {
		dlog("invalid ftp-options");
		return -1;
	}
	new_srvp->variant = flag;
	return 0;
}

int
vol_set_ufs_directory(topdir)
char *topdir;
{
	/*
	 * "topdir" can be either absolute or relative pathname.
	 * Relative pathname can be used to create volumes beneath
	 * WWFSDIR (assuming csd has chdir()ed there).
	 */
	if (validate_dir(topdir) < 0) {
		dlog("ufs-directory %s: %m", topdir);
		return -1;
	}
	new_volp->flag |= WF_VOL_DIRECTFILE;
	new_volp->topdir = strdup(topdir);
	new_srvp = (wf_srv *)0;
	if (MORE(&new_volp->srvq->q)) {
		/* XXX more cleaner implementation desired but, for now, */
		dlog("ufs-directory supercedes other server descriptions");
	}
	return 0;
}

int
vol_set_description(desc)
char *desc;
{
	/* simply ignore it for now */
	return 0;
}

static wf_tailor vol_tailor[] = {
	{ "volume-id", vol_set_id },
	{ "volume-name", vol_set_name },
	{ "ftp-server", vol_set_ftp_server },
	{ "ftp-directory", vol_set_ftp_directory },
	{ "ftp-options", vol_set_ftp_options },
	{ "ufs-directory", vol_set_ufs_directory },
	{ "description", vol_set_description },
	/*
	 * just for backward compatibility...
	 */
	{ "server", vol_set_ftp_server },
	{ "directory", vol_set_ftp_directory },
	{ NULL, (int (*)())0 }
};

int
vol_bind(name, value)
char *name;
char *value;
{
	return parse_tailor(vol_tailor, name, value);
}

wf_vol *
vol_load(dir, name)
char *dir, *name;
{
	char path[MAXPATHLEN];
	struct stat statbuf;

	sprintf(path, "%s/vol/%s/%s.vol", cs_topdir, dir, name);
	if (access(path, R_OK) < 0)
		return (wf_vol *)0;
	stat(path, &statbuf);
	new_srvp = (wf_srv *)0;
	dir_count = srv_count = 0;
	new_volp = vol_alloc();
	new_volp->dir = strdup(dir);
	new_volp->name = strdup(name);
	new_volp->accp = acc_findname(dir);
	new_volp->mtime = statbuf.st_mtime;

	if (parse_conf(path, vol_bind) < 0) {
		vol_free(new_volp);
		return (wf_vol *)0;
	}
	if (new_volp->id == 0) {
		dlog("%s: no volume id", path);
		vol_free(new_volp);
		return (wf_vol *)0;
	}
	return new_volp;
}

void
vol_save(volp)
wf_vol *volp;
{
	FILE *fp;
	wf_srvq *srvp;
	char path[MAXPATHLEN];

	sprintf(path, "%s/vol/%s.vol", cs_topdir, volp->name);
	mkdirs(path);
	fp = fopen(path, "w");
	if (! fp) return;

	/* write per volume info */
	fprintf(fp, "volume-id:\t%d\n", volp->id);

	/* write server names */
	ITER(srvp, wf_srvq, &volp->srvq->q) {
		/* XXX provision for other access types required */
		fprintf(fp, "ftp-server:\t%s\n", srvp->srv->name);
		fprintf(fp, "ftp-directory:\t%s\n", srvp->info);
	}
	fclose(fp);
}

wf_conn *
vol_getconn(volp)
wf_vol *volp;
{
	wf_conn *cp;
	wf_srvq *srvq;

	if (volp->nconn < WF_VOL_MAXCONN) {
		volp->flag &= ~WF_VOL_XCONN;
		srvq = pol_select_srv(volp);
		cp = srv_getconn(srvq);
		if (cp) {
			++volp->nconn;
			return cp;
		}
	} else {
		volp->flag |= WF_VOL_XCONN;
	}
	return (wf_conn *)0;
}

wf_conn *
vol_newconn(volp)
wf_vol *volp;
{
	wf_conn *cp;
	wf_srvq *srvq;

	if (volp->nconn < WF_VOL_MAXCONN) {
		volp->flag &= ~WF_VOL_XCONN;
		srvq = pol_select_srv(volp);
		cp = srv_newconn(srvq);
		if (cp) {
			++volp->nconn;
			return cp;
		}
	}
	volp->flag |= WF_VOL_XCONN;
	return (wf_conn *)0;
}

void
vol_releaseconn(volp)
wf_vol *volp;
{
	warn_negative(--volp->nconn);
}

void
vol_addsrv(volp, srvp, subdir)
wf_vol *volp;
wf_srv *srvp;
char *subdir;
{
	wf_srvq *q = ALLOC(wf_srvq);
	q->srv = srvp;
	q->info = (void *)subdir;
	q_insert(q, LAST(wf_srvq, &volp->srvq->q));
}

void
vol_deletesrv(volp, srvp)
wf_vol *volp;
wf_srv *srvp;
{
	wf_srvq *q, *q2;
	qelem *q_head = &volp->srvq->q;

	ITER2 (q, q2, wf_srvq, q_head) {
		if (q->srv == srvp) {
			q_remove(q);
			free(q->info);
			FREE(q);
			return;
		}
	}
	syslog(LOG_WARNING, "vol_deletesrv: should not reach here");
}

long
vol_allocid(volp)
wf_vol *volp;
{
	/* Allocate a new directory id.
	 *
	 * NFS Version 2 Protocol Specification says:
	 * "fsid" is the file system identifier for the filesystem
	 * containing the file.
	 * "fileid" is a number that uniquely identifies the file
	 * within its file system.
	 *
	 * I think we satisfy this requirement by assigning a unique
	 * volume id to "fsid" and a file id to "fileid". Each fileid
	 * is guaranteed to be unique within given fsid.
	 */

#define	MAPSIZE	256
	/* 256 came from "blocksize / sizeof(long)", which is sufficient
	 * I believe, since each volume can contain
	 * 256(longs) * 32(bits/long) = 8192 directories */

	int fd;
	unsigned long bitmap[MAPSIZE];
	long i, j;
	char path[MAXPATHLEN];

	/*
	 * load bitmap
	 */
	/* XXX tuning is possible for contiguous allocid */
	sprintf(path, "%s/cache/%lx/dir/bitmap", cs_topdir, volp->id);
	if (access(path, R_OK|W_OK) != 0) {
		if (mkdirs(path))
			dir_realdir(path, volp, (wf_dir *)0);
		fd = open(path, O_CREAT|O_TRUNC|O_RDWR, 0660);
		memset(bitmap, 0xff, MAPSIZE * sizeof(long));
	} else {
		fd = open(path, O_RDWR, 0660);
		read(fd, bitmap, MAPSIZE * sizeof(long));
	}

	/*
	 * find unused bit and mark it
	 */
	for (i = 0; bitmap[i] == 0; ++i) ;
	j = ffs(bitmap[i]) - 1;
	bitmap[i] &= ~(1 << j);

	/*
	 * save bitmap 
	 */
	/* XXX tuning is possible for contiguous allocid */
	lseek(fd, 0L, 0);
	write(fd, bitmap, MAPSIZE * sizeof(long));
	close(fd);

	/*
	 * return allocated id
	 */
	return (i * 8 * sizeof(long) + j + 1);

}

void
vol_destroyid(volp)
wf_vol *volp;
{
	char path[MAXPATHLEN];

	sprintf(path, "%s/cache/%lx/dir/bitmap", cs_topdir, volp->id);
	unlink(path);
	if (access(path, R_OK) == 0) {
		dlog("could not unlink %s", path);
	}
}

wf_dir *
vol_newroot(volp)
wf_vol *volp;
{
	wf_dir *dirp;
	long dirid;
	int retry_count = 0;
	char path[MAXPATHLEN], path2[MAXPATHLEN];

 retry:
	dirid = vol_allocid(volp);
	if (dirid != 1) {
		if (++retry_count > 3) {
			/* something is wrong.. */
			errno_diag();
			return (wf_dir *)0;
		}
		/* Oops. Possible causes are:
		 * - ID bitmap of this volume has been corrupted.
		 * - Someone deleted root directory cache (".../%lx/%lx/1").
		 */
		dlog("Directory cache of volume %s deleted?", volp->name);
		sprintf(path2, "%s/stale/%s.%d/%lx/",
			cs_topdir, cs_date, cs_serial, volp->id);
		mkdirs(path2);
		sprintf(path, "/bin/mv %s/cache/%lx/dir %s",
			cs_topdir, volp->id, path2);
		unix_command(path);
		cs_serial++;
		goto retry;
	}
	/* Directory id 1 is just for sanity check.
	 * We use volume id instead, biased with WF_VOL_BASE, to avoid ID
	 * conflict with other volumes. Otherwise pwd doesn't work correctly.
	 */
	dirp = dir_new(volp, WF_VOL_BASE + volp->id, 0, 0);
	return dirp;
}

wf_dir *
vol_findroot(volp)
wf_vol *volp;
{
	wf_dir *dirp;

	dirp = dir_findcache(volp, WF_VOL_BASE + volp->id);
	return dirp;
}

void
vol_inspect(reqp, name)
wf_req *reqp;
char *name;
{
	int ret;
	wf_ih ih;
	int id = reqp->id;

	ret = root_name2ih(name, &ih);
	if (ret != WF_REP_OK) {
		req_send(reqp, "513.%03d root_name2ih: error code %d.\n",
			 id, ret);
		return;
	}
	req_send(reqp, "220-%03d vol %s\n", id, name);
	req_send(reqp, "213-%03d name = %s\n", id, PROT(ih.child_volp->name));
	req_send(reqp, "213-%03d id = %lx\n", id, ih.child_volp->id);
	req_send(reqp, "213-%03d flag = %x\n", id, ih.child_volp->flag);
	req_send(reqp, "213-%03d nconn = %d\n", id, ih.child_volp->nconn);
	req_send(reqp, "213.%03d mtime = %ld (%20.20s)\n",
		 id, ih.child_volp->mtime, ctime(&ih.child_volp->mtime)+4);
}

