/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/* $Header: wfs.h[109.0] Wed Nov 24 03:47:20 1993 youki-k@is.aist-nara.ac.jp saved $ */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <rpc/rpc.h>
#ifndef IPPROTO_UDP
#include <netinet/in.h>
#endif
#include <arpa/inet.h>
#include "nfs_prot.h"
#include "cs_prot.h"
#define WFS_H
#include "parse.h"

typedef enum {
	Start,
	Run,
	Finishing,
	Quit,
	Done
} serv_state;

typedef struct sockaddr_in sockaddr_in;
typedef void (*func)();
typedef struct svc_req svc_req;

/*
 * Linked list
 */
typedef struct qelem qelem;
struct qelem {
	qelem *q_forw;
	qelem *q_back;
};
#define FIRST(ty, q)	((ty *) ((q)->q_forw))
#define LAST(ty, q)	((ty *) ((q)->q_back))
#define NEXT(ty, q)	((ty *) (((qelem *) q)->q_forw))
#define PREV(ty, q)	((ty *) (((qelem *) q)->q_back))
#define HEAD(ty, q)	((ty *) q)
#define MORE(q)		((q)->q_forw != q)
#define ITER(v, ty, q) \
	for ((v) = FIRST(ty,(q)); (v) != HEAD(ty,(q)); (v) = NEXT(ty,(v)))
#define REVITER(v, ty, q) \
	for ((v) = LAST(ty,(q)); (v) != HEAD(ty,(q)); (v) = PREV(ty,(v)))
#define ITER2(v, w, ty, q) \
	for ((v) = FIRST(ty, (q)); (w) = NEXT(ty, (v)), (v) != HEAD(ty, (q)); \
	     (v) = (w))

/* protocol switch & protocol spec */
typedef struct wf_proto {
	char		*name;		/* protocol name */
	func		wf_getattr;
	func		wf_getattr_miss;
	func		wf_lookup;
	func		wf_lookup_miss;
	func		wf_readlink;
	func		wf_read;
	func		wf_read_miss;
	func		wf_readdir;
	func		wf_readdir_miss;
	func		wf_getuda;	/* user defined attributes */
	func		wf_keepalive;
	func		wf_close;
	func		wf_shutdown;
	int		class;		/* protocol class, WF_CLASS */
} wf_proto;

/* wf_proto->class bitmap */
#define WF_CLASS_AUTH	0x1	/* authentication */
#define WF_CLASS_ALIVE	0x2	/* keepalive */
#define WF_CLASS_RD	0x4	/* reliable delivery */

/* to compute padding */
struct wf_fhsize {
	long	a;
	long	b;
	long	c;
	long	d;
	long	e;
};

/* file handle */
typedef struct wf_fh {
	long	world_id;	/* filehandle type and version */
	long	parent_vol;	/* unique in the world */
	long	child_vol;	/* unique in the world */
	union {
		struct {
			long	u_dir_id;	/* inode */
			long	u_file_id;	/* crc */
		} u_file;
		struct {
			long	u_parent_dir;	/* inode# of parent dir */
			long	u_child_dir;	/* inode# of child dir */
		} u_dir;
	} u;
	char	padding[NFS_FHSIZE - sizeof(struct wf_fhsize)];
} wf_fh;

#define	dir_id		u.u_file.u_dir_id
#define	file_id		u.u_file.u_file_id
#define	parent_dir	u.u_dir.u_parent_dir
#define	child_dir	u.u_dir.u_child_dir

typedef struct wf_srvstat {
	long	bytes_to_server;	/* # of bytes sent to servers */
	long	bytes_from_server;	/* # of bytes received from servers */
	long	rtt;			/* round trip time with decay */
	long	icmp_reply;		/* # of successful echo-reply */ 
	long	icmp_unreach;		/* # of dest unreach */
} wf_srvstat;

typedef struct wf_connq wf_connq;
/* per-server information */
typedef struct wf_srv {
	qelem		q;
	char		*name;		/* server name */
	sockaddr_in	sin;		/* network address */
	int		flag;		/* server status, WF_SRV */
	long		touch;		/* time of last status change */
	int		protos;		/* supported protocols, WF_PROTO */
	long		variant;	/* extension flags, WF_XYZFS */
	wf_connq	*connq;		/* connections for this server */
	int		nconn;		/* number of connections */
	wf_srvstat	stats;		/* per-server statistics */
} wf_srv;

/* per-volume-server information */
typedef struct wf_srvq {
	qelem	q;
	wf_srv	*srv;
	void	*info;		/* protocol specific */
} wf_srvq;

/* wf_srv->flag */
#define	WF_SRV_HUNREACH	0x1	/* server unreachable */
#define	WF_SRV_NUNREACH	0x2	/* network unreachable */
#define	WF_SRV_REFUSED	0x4	/* connection refused */
#define	WF_SRV_XABORT	0x8	/* transfer aborted */
#define	WF_SRV_TIMEOUT	0x10	/* transfer timeout */
#define	WF_SRV_BADNS	0x20	/* name-server problem */

/* wf_srv->protos bitmap */
#define WF_PROTO_FTP	0x1	/* file transfer protocol */
#define WF_PROTO_NFSUDP	0x2	/* sun NFS over UDP */
#define WF_PROTO_NFSTCP	0x4	/* sun NFS over TCP */
#define WF_PROTO_YFS	0x8	/* youki file system */

/* wf_srv->variant bitmap */
#define WF_FTP_FOLLOW	0x1	/* follow symbolic links */
#define WF_FTP_NOSTAT	0x2	/* FTP does not support STAT */
#define WF_NFS_SPONGY	0x4	/* NFS spongy mount */
#define WF_NFS_COMPRESS	0x8	/* NFS compress */

#define	WF_SRV_MAXCONN	16		/* max # of connection for a server */

typedef struct wf_connstat {
	long	bytes_to_server;	/* # of bytes sent to servers */
	long	bytes_from_server;	/* # of bytes received from servers */
} wf_connstat;

typedef struct wf_thrd wf_thrd;
/* per-connection information (server side) */
typedef struct wf_conn {
	qelem		q;
	wf_proto	*proto;		/* protocol used */
	int		id;		/* connection id */
	wf_thrd		*thrdp;		/* transfer */
	int		idle;		/* idle time */
	wf_srv		*srv;		/* peer entity */
	wf_connstat	stats;		/* per-connection statistics */
	void		*info;		/* protocol specific */
} wf_conn;

/* list of per-connection information */
struct wf_connq {
	qelem	q;
	wf_conn	*conn;
};

/* per-session information (client side); primarily for non-RPC protocols */
typedef struct wf_req {
	long		id;		/* session ID */
	int		so;		/* socket */
	sockaddr_in	sin;		/* client address */
	int		flag;		/* flags */
} wf_req;

#define	WF_REQ_PRIV	0x01		/* privileged connection */

/* access control information */
typedef struct wf_acc {
	qelem		q;
	unsigned long	from;		/* lower bound (host byte order) */
	unsigned long	to;		/* upper bound (host byte order) */
	long		flag;		/* WF_ACC_ */
} wf_acc;

#define	WF_ACC_NFS	0x1
#define	WF_ACC_UIP	0x2
#define	WF_ACC_FDP	0x4
#define	WF_ACC_RAW	0x8

typedef struct wf_volstat {
	long	bytes_to_server;	/* # of bytes sent to servers */
	long	bytes_from_server;	/* # of bytes received from servers */
	long	bytes_to_client;	/* # of bytes sent to clients */
	long	bytes_from_client;	/* # of bytes received from clients */
	long	n_request_readdir;	/* # of readdir requests */
	long	n_request_lookup;	/* # of lookup requests */
	long	n_request_read;		/* # of read requests */
	long	n_request_readdir_miss;	/* # of readdir actually done */
	long	n_request_lookup_miss;	/* # of lookup actually done */
	long	n_request_read_miss;	/* # of read actually done */
} wf_volstat;

typedef struct wf_dir wf_dir;
/* list of per-volume information */
typedef struct wf_vol {
	qelem		q;
	char		*name;		/* volume name */
	long		id;		/* volume id */
	char		*dir;		/* directory this volume resides */
	wf_acc		*accp;		/*   and its access control info */
	int		flag:16;	/* status */
	int		nconn:16;	/* number of connections */
	long		mtime;		/* timestamp of volume update */
	wf_srvq		*srvq;		/* servers of this volume */
	wf_dir		*dirp;		/* cached directory chain */
	wf_thrd		*thrdp;		/* transfer */
	char		*topdir;	/* for WF_VOL_DIRECTFILE */
	wf_volstat	stats;		/* per-volume statistics */
} wf_vol;

/* wf_vol->flag */
#define	WF_VOL_XCONN		0x1	/* exceeded connection limit */
#define	WF_VOL_DIRECTFILE	0x2	/* bypass file caching mechanism */

/* constants */
#define	WF_VOL_MAXCONN	4	/* max # of connection for a volume */
#define	WF_VOL_BASE	0x10000	/* large enough so that volume IDs can be
				 * discriminated from directory ID */
#define	WF_GLUE_BASE	0x8000	/* large enough so that glue IDs can be
				 * discriminated from directory ID */
#define	WF_ROOT_VOL_ID	0	/* root file system */
#if 0
#define	WF_HOST_VOL_ID	(-1)	/* ".wwfs_host" */
#define	WF_CWD_VOL_ID	(-2)	/* ".wwfs_cwd" */
#endif
#define	WF_FH_VOL_ID	(-3)	/* ".wwfs_fh" */

/* list of per-file information */
typedef struct wf_file {
	qelem	q;
	char	*name;		/* filename */
	char	*link;		/* symbolic link */
	long	id;		/* file id (crc32 or inode #) */
	long	crc;		/* crc32 */
	int	flag;		/* WF_FILE */
	struct fattr fattr;	/* see nfs_prot.h */
	wf_thrd	*thrdp;		/* transfer */
} wf_file;

/* wf_file->flag */
#define WF_FILE_MARKED		0x1
#define	WF_FILE_MUTABLE		0x2

/* wrapper macros for this complex data structure */
#define	FSID(p)		((p)->fattr.fsid)
#define	FID(p)		((p)->id)

/* list of per-directory information */
struct wf_dir {
	qelem	q;
	char	*name;		/* entire directory name */
	long	id;		/* directory id (inode #) */
	int	flag;		/* WF_DIR */
	int	idle;		/* idle time */
	long	ctime;		/* when this id was allocated */
	wf_file	*file;		/* files in this directory */
	int	n_xfer;		/* # of ongoing transfer */
};

#define	WF_ROOT_DIR_ID	0xffff

/* wf_dir->flag */
#define	WF_DIR_OBSOLETE		0x1
#define	WF_DIR_WIREDDOWN	0x2

typedef struct wf_ih {		/* internal handle */
	wf_vol	*parent_volp;
	wf_dir	*dirp;
	wf_vol	*child_volp;
	wf_file	*filep;
} wf_ih;

/* for filesystem trace */
typedef struct wf_trace {
	qelem	q;
	long	cookie;		/* unique id */
	long	time;		/* timestamp */
	int	uid;		/* user id */
	long	client;		/* client ip address */
	long	volume;		/* volume id */
	long	dir;		/* directory id */
	long	file;		/* file id */
	int	op:16;		/* operation. read(0) or readdir(1) */
	int	hit:16;		/* cache hit(1) or miss(0) */
	long	arg;		/* file size etc. */
} wf_trace;

/* wf_trace->op */
#define	WF_OP_READ		0
#define	WF_OP_READDIR		1
/* wf_trace->hit */
#define	WF_CACHE_MISS		0
#define	WF_CACHE_HIT		1
#define	WF_CACHE_READY		2

typedef struct callout callout;
/* Unified closure.
 * Represents all information about currently processing request.
 */
struct wf_thrd {
	qelem		q;
	int		step:16;	/* state of this thread */
	int		flag:16;	/* WF_THREAD */
	/* nfs args */
	svc_req		*rqstp;
	wf_vol		*parent_volp;
	wf_vol		*child_volp;
	wf_fh		*fhp;
	char		*fname;
	nfscookie	cookie;		/* nfs_readdir */
	u_int		offset;		/* nfs_read */
	u_int		count;		/* nfs_readdir, nfs_read */
	u_int		totalcount;	/* nfs_read */
	/* state */
	wf_fh		*resfhp;
	wf_dir		*dirp;
	wf_dir		*olddirp;
	wf_dir		*dotdotp;
	wf_file		*filep;
	/* for user-interface protocols */
	wf_req		req;		/* deepcopy here */
	char		*pathname;	/* deepcopy full pathname */
	int		uid;		/* user id, for trace */
	unsigned long	client;		/* ip address, for trace */
	/* transfer state */
	wf_proto	*proto;
	wf_conn		*cp;		/* connection for this request */
	int		so;		/* data connection socket */
	int		port;		/* port for data connection */
	int		fd;		/* received data */
	char		*buf;		/* received data */
	int		bytes;		/* done so far */
	int		idle;		/* risky but no response */
	func		reply;		/* reply to client */
	func		give_up;	/* encountered difficulty */
	void		*closure;	/* for give_up */
	callout		*callout;	/* timeout handle (opaque) */
};

/* wf_thrd->flag bitmap */
#define WF_THREAD_BUSY		0x1
#define WF_THREAD_TIMEOUT	0x2	/* 2: timeout, ~2: in time */
#define WF_THREAD_RISKY		0x4	/* 4: risky, ~4: safe */
#define WF_THREAD_ERROR		0x8	/* 8: error, ~8: no error */

/* wf_thrd->step FSM  */
#define	FTPS_PROMPT		0
#define	FTPS_STAT_RECV		1
#define	FTPS_STAT_RECVINIT	2
#define	FTPS_STAT_RECVLOOP	3
#define	FTPS_RETR_SEND		4
#define	FTPS_PORT_ACCEPT	5
#define	FTPS_RETR_RECVINIT	6
#define	FTPS_RETR_RECVLOOP	7
#define	FTPS_RETR_FIN		8
#define	FTPS_CONN_INIT		10
#define	FTPS_CONN_USER		11
#define	FTPS_CONN_PASS		12
#define	FTPS_CONN_TYPE		13
#define	FTPS_CONN_FIN		14
#define	FTPS_CONN_FAIL		15

/* unified error code. limited to internal use. */
#define	WF_REP_OK	0		/* no error */
#define	WF_ERR_PERM	1		/* Not owner */
#define	WF_ERR_NOENT	2		/* No such entity */
#define	WF_ERR_IO	3		/* I/O error */
#define	WF_ERR_NXIO	4		/* No such device or address */
#define	WF_ERR_ACCES	5		/* Permission denied */
#define	WF_ERR_EXIST	6		/* Given resource exists */
#define	WF_ERR_NODEV	7		/* No such device */
#define	WF_ERR_NOTDIR	8		/* Not a directory*/
#define	WF_ERR_ISDIR	9		/* Is a directory */
#define	WF_ERR_FBIG	10		/* Resource too large */
#define	WF_ERR_NOSPC	11		/* No space left on device */
#define	WF_ERR_ROFS	12		/* Read-only file system */
#define	WF_ERR_NAMETOOLONG	13	/* Resource name too long */
#define	WF_ERR_NOTEMPTY	14		/* Directory not empty */
#define	WF_ERR_DQUOT	15		/* Disc quota exceeded */
#define	WF_ERR_STALE	16		/* Stale file handle */
#define	WF_ERR_WFLUSH	17		/* Write cache flushed */
#define	WF_REP_WAIT	18		/* Operation would block */
#define	WF_ERR_ALREADY	19		/* Operation already in progress */
#define	WF_ERR_PROTO	20		/* Protocol not available */
#define	WF_ERR_NDOWN	21		/* Network is down */
#define	WF_ERR_NUNREACH	22		/* Network is unreachable */
#define	WF_ERR_XABORT	23		/* Transfer aborted */
#define	WF_ERR_TIMEOUT	24		/* Connection timed out */
#define	WF_ERR_REFUSED	25		/* Connection refused */
#define	WF_ERR_HDOWN	26		/* Host is down */
#define	WF_ERR_HUNREACH	27		/* No route to host */
#define	WF_ERR_NQUOT	28		/* Network quota exceeded */
