# libmm.pl -- package for accessing csd cache.
# Author: Youki Kadobayashi <youki@wide.ad.jp>
#
# It is written in the hope that perl scripts built upon this package
# complements features not implemented in csd, thereby enabling much
# more advanced features such as cache strategy selection, cache statistics
# and cache consistency management.  I have also added UIP commands that
# enables close collaboration of csd and these perl programs.
#
# This package only works on localhost (that is, the machine running csd)
# for security reasons.  Current implementation of UIP server restricts
# CTOR/RTOC access to localhost.

# You may need to run `h2ph' on `sys/socket.h' to get the proper values
# handy in a perl library file.
require 'sys/socket.ph' || die $!;

package mm;
$S_IFDIR = 040000;	# because most of you don't have 'sys/stat.pl'...
$S_IFREG = 0100000;

#
#	initialization: create UDP socket for UIP.
#
$uip_port = 8002;
$sockaddr = 'S n a4 x8';
@addr = gethostbyname('localhost');
$localip = pop(@addr);
$laddr = pack($sockaddr, &'AF_INET, 0, $localip);
$faddr = pack($sockaddr, &'AF_INET, $uip_port, $localip);
socket(SO, &'AF_INET, &'SOCK_DGRAM, 0) || die $!;
bind(SO, $laddr) || die $!;

$serial = 0;
$debug = 0;			# 1 if you want to debug your code

sub uip_send
{
    local ($msg) = @_;
    send(SO, "$msg\n", 0, $faddr);
    print "$msg\n" if $debug;
}

sub uip_recv
{
    local ($msg, $from);
    do {
	$from = recv(SO, $msg, 256, 0);
	print "$msg" if $debug;
    } until unpack("x4 a4 x8", $from) eq $localip;
    return $msg;
}

#
#	mm'fstoc: convert from real (filesystem) pathname to cache pathname.
#	static.
#
sub fstoc
{
    local ($path, $type) = @_;
    local ($cachepath, $msg);

    if ($type == $S_IFDIR) {
	# convert it to the pathname of a cached directory information,
	# namely, "%s/%lx/dir/%lx"
	&uip_send("RTOC dir $path");
    }
    elsif ($type == $S_IFREG) {
	# convert it to the pathname of a cached file,
	# namely, "%s/%lx/%lx/%s"
	&uip_send("RTOC fil $path");
    }
    else {
	$error = "fstoc: unsupported file type: $type\n";
	return undef;
    }

    while (1) {
	$msg = &uip_recv;
	if ($msg =~ /^2..\.... c_p /) {
	    # we got cache path
	    $cachepath = $';
	    chop $cachepath;
	    return ($cachepath);	# unique pathname can be the key
	}
	elsif ($msg =~ /^[45]..\.... /) {
	    $error = $';
	    return undef;
	}
    }
}

sub dir_file
{
    local ($path) = @_;
    local ($dir, $file, @components);

    @components = split(m:/+:, $path);
    $file = pop(@components);
    $dir = join('/', @components);
    return ($dir, $file);
}

#
#	mm'ctofs: convert from cache pathname to real (filesystem) pathname.
#	static.
#
sub ctofs
{
    local ($path) = @_;
    local (@res, $realpath);
    local ($dir, $file);

    @res = stat($path);
    if ($res[2] & $S_IFDIR) {
	open(IN, "< $path/.realdir") || return undef;
	$realpath = <IN>; close(IN);
	return $realpath;
    }
    elsif ($res[2] & $S_IFREG) {
	($dir, $file) = &dir_file($path);
	open(IN, "$dir/.realdir") || return undef;
	$realpath = <IN>; close(IN);
	return "$realpath/$file";
    }
    else {
	$error = "ctofs: no such file or directory: $path\n";
	return undef;
    }
}

#
#	&mm'opendir(*DIRHANDLE, pathname)
#
sub opendir
{
    local (*DIR, $pathname) = @_;
    local ($cachepath, $key, $size);
    local ($magic, $namelen, $name, $ctime, $nfile);

    $cachepath = &fstoc($pathname, $S_IFDIR) || return undef;
    open(DIR, "<$cachepath") || return undef;
    $key = fileno(DIR);		# use file number as the unique key
    $size = (stat(DIR))[7];
    $list{$key} = "";		# otherwise bad address occur
    read(DIR, $list{$key}, $size) || return undef;

    ($namelen) = unpack("x4 i", $list{$key});
    ($magic, $name, $ctime, $nfile) = unpack("l x4 A$namelen l i",$list{$key});
    $name{$key} = $name;
    $ctime{$key} = $ctime;
    $nfile{$key} = $nfile;
    $index{$key} = 0;
    $offset{$key} = 4 * 4 + $namelen;
    1;
}

#
#	&mm'closedir(*DIRHANDLE)
#
sub closedir
{
    local (*DIR) = @_;
    local ($key) = fileno(DIR);

    close(DIR);
    # conserve memory...
    undef($list{$key});
    undef($name{$key});
    undef($ctime{$key});
    undef($nfile{$key});
    undef($index{$key});
    undef($offset{$key});
    0;
}

#
#	($name, $link, $type, $crc, $ausec, $musec, $cusec,
#	 $fsid, $fid, $mode, $nlink, $uid, $gid, $rdev, $size,
#	 $asec, $msec, $csec, $blocksize, $blocks)
#		= &mm'getnext(*DIRHANDLE)
#
#	getnext: readdir + stat complex.  Return next directory entry.
#	mm'opendir must be called before getnext.
#
sub getnext
{
    local (*DIR) = @_;
    local ($key) = fileno(DIR);
    local ($namelen, $linklen, $name, $link);
    local ($fid, $type, $mode, $nlink, $uid, $gid,
	   $size, $blocksize, $rdev, $blocks, $fsid, $crc,
	   $asec, $ausec, $msec, $musec, $csec, $cusec);
    local ($off) = $offset{$key};

    die "getnext: directory not opened\n" if !defined($name{$key});
    return () if (++$index{$key} >= $nfile{$key});
    ($fid, $namelen) = unpack("x$off l l", $list{$key});
    ($name) = unpack("x$off x8 A$namelen", $list{$key});
    ($linklen) = unpack("x$off x8 x$namelen l", $list{$key});
    ($link) = unpack("x$off x8 x$namelen x4 A$linklen", $list{$key});
    $off += 12 + $namelen + $linklen;
    ($type, $mode, $nlink, 
     $uid, $gid, $size, $blocksize, $rdev, $blocks, $fsid, $crc,
     $asec, $ausec, $msec, $musec, $csec, $cusec)
	= unpack("x$off l l l l l l l l l l l l l l l l l", $list{$key});
    $off += 4 * 17;
    $offset{$key} = $off;

    return ($name, $link, $type, $crc, $ausec, $musec, $cusec,
	    $fsid, $fid, $mode, $nlink, $uid, $gid, $rdev, $size,
	    $asec, $msec, $csec, $blocksize, $blocks)
}

#
#	$name = &mm'readdir(*DIRHANDLE)
#	@names = &mm'readdir(*DIRHANDLE)
#
#	In array context, return list of filenames.
#	In scalar context, return next directory entry.
#
sub readdir
{
    local (*DIR) = @_;
    local ($key) = fileno(DIR);
    local (@files, $i);

    die "readdir: directory not opened\n" if !defined($name{$key});
    if (wantarray) {
	for ($i = 0; $i < $nfile{$key}; ++$i) {
	    push(@files, (&getnext(*DIR))[0]);
	}
	return @files;
    }
    return (&getnext(*DIR))[0];
}

#
#	rewinddir, seekdir, telldir is not implemented yet.
#	do you need it?
#

#
#	($fsid, $fid, $mode, $nlink, $uid, $gid, $rdev, $size,
#	 $asec, $msec, $csec, $blocksize, $blocks)
#		= &mm'stat(*DIRHANDLE, $filename)
#
#	Stat call is heavyweight in this library.  Not recommended.
#	As a penalty, I made it incompatible with stat system call.
#
sub stat
{
    local (*DIR, $filename) = @_;
    local ($key) = fileno(DIR);
    local (@res, $i);

    for ($i = 0; $i < $nfile{$key}; ++$i) {
	@res = &getnext(*DIR);
	if ($res[0] eq $filename) {
	    return @res[7 .. $#res];
	}
    }
    -1;		# errno = ENOENT
}

#
#	$status = &mm'open(FILEHANDLE, EXPR)
#
sub open
{
    local (*FH, $expr) = @_;
    local ($key, $pos, $flag, $pathname);

    $pos = index($expr, "/");
    if ($pos < 0) {
	open(FH, $expr) || return undef;
	return 1;
    }
    ($flag, $pathname) = unpack("a$pos a*", $expr);
    $key = &fstoc($pathname, $S_IFREG) || return undef;
    open(FH, "$flag$key") || return undef;
    return 1;
}

#
#	There are no counterpart for built-in functions like
#	read(), write(), seek() and close(); they just work fine.
#

1;
